@@ -37,4 +37,11 @@ pq_peek_items(poe_queue *pq, SV *filter, int max_count, pq_entry **items);
 extern void pq_dump(poe_queue *pq);
 extern void pq_verify(poe_queue *pq);
 
+#ifdef USE_ITHREADS
+extern poe_queue *pq_clone(poe_queue *pq);
+#endif
+
+extern void
+pq__set_errno_queue(int value);
+
 #endif
